# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keras_tuner/protos/service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from keras_tuner.protos import keras_tuner_pb2 as keras__tuner_dot_protos_dot_keras__tuner__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='keras_tuner/protos/service.proto',
  package='keras_tuner',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n keras_tuner/protos/service.proto\x12\x0bkeras_tuner\x1a$keras_tuner/protos/keras_tuner.proto\"\x11\n\x0fGetSpaceRequest\"I\n\x10GetSpaceResponse\x12\x35\n\x0fhyperparameters\x18\x01 \x01(\x0b\x32\x1c.keras_tuner.HyperParameters\"K\n\x12UpdateSpaceRequest\x12\x35\n\x0fhyperparameters\x18\x01 \x01(\x0b\x32\x1c.keras_tuner.HyperParameters\"\x15\n\x13UpdateSpaceResponse\"&\n\x12\x43reateTrialRequest\x12\x10\n\x08tuner_id\x18\x01 \x01(\t\"8\n\x13\x43reateTrialResponse\x12!\n\x05trial\x18\x01 \x01(\x0b\x32\x12.keras_tuner.Trial\"\xa3\x01\n\x12UpdateTrialRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12=\n\x07metrics\x18\x02 \x03(\x0b\x32,.keras_tuner.UpdateTrialRequest.MetricsEntry\x12\x0c\n\x04step\x18\x03 \x01(\x03\x1a.\n\x0cMetricsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01\"?\n\x13UpdateTrialResponse\x12(\n\x06status\x18\x01 \x01(\x0e\x32\x18.keras_tuner.TrialStatus\"M\n\x0f\x45ndTrialRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12(\n\x06status\x18\x02 \x01(\x0e\x32\x18.keras_tuner.TrialStatus\"\x12\n\x10\x45ndTrialResponse\"*\n\x14GetBestTrialsRequest\x12\x12\n\nnum_trials\x18\x01 \x01(\x03\";\n\x15GetBestTrialsResponse\x12\"\n\x06trials\x18\x01 \x03(\x0b\x32\x12.keras_tuner.Trial\"#\n\x0fGetTrialRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t\"5\n\x10GetTrialResponse\x12!\n\x05trial\x18\x01 \x01(\x0b\x32\x12.keras_tuner.Trial2\xbf\x04\n\x06Oracle\x12I\n\x08GetSpace\x12\x1c.keras_tuner.GetSpaceRequest\x1a\x1d.keras_tuner.GetSpaceResponse\"\x00\x12R\n\x0bUpdateSpace\x12\x1f.keras_tuner.UpdateSpaceRequest\x1a .keras_tuner.UpdateSpaceResponse\"\x00\x12R\n\x0b\x43reateTrial\x12\x1f.keras_tuner.CreateTrialRequest\x1a .keras_tuner.CreateTrialResponse\"\x00\x12R\n\x0bUpdateTrial\x12\x1f.keras_tuner.UpdateTrialRequest\x1a .keras_tuner.UpdateTrialResponse\"\x00\x12I\n\x08\x45ndTrial\x12\x1c.keras_tuner.EndTrialRequest\x1a\x1d.keras_tuner.EndTrialResponse\"\x00\x12X\n\rGetBestTrials\x12!.keras_tuner.GetBestTrialsRequest\x1a\".keras_tuner.GetBestTrialsResponse\"\x00\x12I\n\x08GetTrial\x12\x1c.keras_tuner.GetTrialRequest\x1a\x1d.keras_tuner.GetTrialResponse\"\x00\x62\x06proto3'
  ,
  dependencies=[keras__tuner_dot_protos_dot_keras__tuner__pb2.DESCRIPTOR,])




_GETSPACEREQUEST = _descriptor.Descriptor(
  name='GetSpaceRequest',
  full_name='keras_tuner.GetSpaceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=87,
  serialized_end=104,
)


_GETSPACERESPONSE = _descriptor.Descriptor(
  name='GetSpaceResponse',
  full_name='keras_tuner.GetSpaceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hyperparameters', full_name='keras_tuner.GetSpaceResponse.hyperparameters', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=106,
  serialized_end=179,
)


_UPDATESPACEREQUEST = _descriptor.Descriptor(
  name='UpdateSpaceRequest',
  full_name='keras_tuner.UpdateSpaceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hyperparameters', full_name='keras_tuner.UpdateSpaceRequest.hyperparameters', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=181,
  serialized_end=256,
)


_UPDATESPACERESPONSE = _descriptor.Descriptor(
  name='UpdateSpaceResponse',
  full_name='keras_tuner.UpdateSpaceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=258,
  serialized_end=279,
)


_CREATETRIALREQUEST = _descriptor.Descriptor(
  name='CreateTrialRequest',
  full_name='keras_tuner.CreateTrialRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tuner_id', full_name='keras_tuner.CreateTrialRequest.tuner_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=281,
  serialized_end=319,
)


_CREATETRIALRESPONSE = _descriptor.Descriptor(
  name='CreateTrialResponse',
  full_name='keras_tuner.CreateTrialResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial', full_name='keras_tuner.CreateTrialResponse.trial', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=377,
)


_UPDATETRIALREQUEST_METRICSENTRY = _descriptor.Descriptor(
  name='MetricsEntry',
  full_name='keras_tuner.UpdateTrialRequest.MetricsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='keras_tuner.UpdateTrialRequest.MetricsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='keras_tuner.UpdateTrialRequest.MetricsEntry.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=497,
  serialized_end=543,
)

_UPDATETRIALREQUEST = _descriptor.Descriptor(
  name='UpdateTrialRequest',
  full_name='keras_tuner.UpdateTrialRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='keras_tuner.UpdateTrialRequest.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='keras_tuner.UpdateTrialRequest.metrics', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='step', full_name='keras_tuner.UpdateTrialRequest.step', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_UPDATETRIALREQUEST_METRICSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=380,
  serialized_end=543,
)


_UPDATETRIALRESPONSE = _descriptor.Descriptor(
  name='UpdateTrialResponse',
  full_name='keras_tuner.UpdateTrialResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='keras_tuner.UpdateTrialResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=545,
  serialized_end=608,
)


_ENDTRIALREQUEST = _descriptor.Descriptor(
  name='EndTrialRequest',
  full_name='keras_tuner.EndTrialRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='keras_tuner.EndTrialRequest.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='keras_tuner.EndTrialRequest.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=610,
  serialized_end=687,
)


_ENDTRIALRESPONSE = _descriptor.Descriptor(
  name='EndTrialResponse',
  full_name='keras_tuner.EndTrialResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=689,
  serialized_end=707,
)


_GETBESTTRIALSREQUEST = _descriptor.Descriptor(
  name='GetBestTrialsRequest',
  full_name='keras_tuner.GetBestTrialsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_trials', full_name='keras_tuner.GetBestTrialsRequest.num_trials', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=709,
  serialized_end=751,
)


_GETBESTTRIALSRESPONSE = _descriptor.Descriptor(
  name='GetBestTrialsResponse',
  full_name='keras_tuner.GetBestTrialsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trials', full_name='keras_tuner.GetBestTrialsResponse.trials', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=753,
  serialized_end=812,
)


_GETTRIALREQUEST = _descriptor.Descriptor(
  name='GetTrialRequest',
  full_name='keras_tuner.GetTrialRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='keras_tuner.GetTrialRequest.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=814,
  serialized_end=849,
)


_GETTRIALRESPONSE = _descriptor.Descriptor(
  name='GetTrialResponse',
  full_name='keras_tuner.GetTrialResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial', full_name='keras_tuner.GetTrialResponse.trial', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=851,
  serialized_end=904,
)

_GETSPACERESPONSE.fields_by_name['hyperparameters'].message_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._HYPERPARAMETERS
_UPDATESPACEREQUEST.fields_by_name['hyperparameters'].message_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._HYPERPARAMETERS
_CREATETRIALRESPONSE.fields_by_name['trial'].message_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._TRIAL
_UPDATETRIALREQUEST_METRICSENTRY.containing_type = _UPDATETRIALREQUEST
_UPDATETRIALREQUEST.fields_by_name['metrics'].message_type = _UPDATETRIALREQUEST_METRICSENTRY
_UPDATETRIALRESPONSE.fields_by_name['status'].enum_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._TRIALSTATUS
_ENDTRIALREQUEST.fields_by_name['status'].enum_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._TRIALSTATUS
_GETBESTTRIALSRESPONSE.fields_by_name['trials'].message_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._TRIAL
_GETTRIALRESPONSE.fields_by_name['trial'].message_type = keras__tuner_dot_protos_dot_keras__tuner__pb2._TRIAL
DESCRIPTOR.message_types_by_name['GetSpaceRequest'] = _GETSPACEREQUEST
DESCRIPTOR.message_types_by_name['GetSpaceResponse'] = _GETSPACERESPONSE
DESCRIPTOR.message_types_by_name['UpdateSpaceRequest'] = _UPDATESPACEREQUEST
DESCRIPTOR.message_types_by_name['UpdateSpaceResponse'] = _UPDATESPACERESPONSE
DESCRIPTOR.message_types_by_name['CreateTrialRequest'] = _CREATETRIALREQUEST
DESCRIPTOR.message_types_by_name['CreateTrialResponse'] = _CREATETRIALRESPONSE
DESCRIPTOR.message_types_by_name['UpdateTrialRequest'] = _UPDATETRIALREQUEST
DESCRIPTOR.message_types_by_name['UpdateTrialResponse'] = _UPDATETRIALRESPONSE
DESCRIPTOR.message_types_by_name['EndTrialRequest'] = _ENDTRIALREQUEST
DESCRIPTOR.message_types_by_name['EndTrialResponse'] = _ENDTRIALRESPONSE
DESCRIPTOR.message_types_by_name['GetBestTrialsRequest'] = _GETBESTTRIALSREQUEST
DESCRIPTOR.message_types_by_name['GetBestTrialsResponse'] = _GETBESTTRIALSRESPONSE
DESCRIPTOR.message_types_by_name['GetTrialRequest'] = _GETTRIALREQUEST
DESCRIPTOR.message_types_by_name['GetTrialResponse'] = _GETTRIALRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetSpaceRequest = _reflection.GeneratedProtocolMessageType('GetSpaceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSPACEREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.GetSpaceRequest)
  })
_sym_db.RegisterMessage(GetSpaceRequest)

GetSpaceResponse = _reflection.GeneratedProtocolMessageType('GetSpaceResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSPACERESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.GetSpaceResponse)
  })
_sym_db.RegisterMessage(GetSpaceResponse)

UpdateSpaceRequest = _reflection.GeneratedProtocolMessageType('UpdateSpaceRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESPACEREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.UpdateSpaceRequest)
  })
_sym_db.RegisterMessage(UpdateSpaceRequest)

UpdateSpaceResponse = _reflection.GeneratedProtocolMessageType('UpdateSpaceResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESPACERESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.UpdateSpaceResponse)
  })
_sym_db.RegisterMessage(UpdateSpaceResponse)

CreateTrialRequest = _reflection.GeneratedProtocolMessageType('CreateTrialRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATETRIALREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.CreateTrialRequest)
  })
_sym_db.RegisterMessage(CreateTrialRequest)

CreateTrialResponse = _reflection.GeneratedProtocolMessageType('CreateTrialResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATETRIALRESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.CreateTrialResponse)
  })
_sym_db.RegisterMessage(CreateTrialResponse)

UpdateTrialRequest = _reflection.GeneratedProtocolMessageType('UpdateTrialRequest', (_message.Message,), {

  'MetricsEntry' : _reflection.GeneratedProtocolMessageType('MetricsEntry', (_message.Message,), {
    'DESCRIPTOR' : _UPDATETRIALREQUEST_METRICSENTRY,
    '__module__' : 'keras_tuner.protos.service_pb2'
    # @@protoc_insertion_point(class_scope:keras_tuner.UpdateTrialRequest.MetricsEntry)
    })
  ,
  'DESCRIPTOR' : _UPDATETRIALREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.UpdateTrialRequest)
  })
_sym_db.RegisterMessage(UpdateTrialRequest)
_sym_db.RegisterMessage(UpdateTrialRequest.MetricsEntry)

UpdateTrialResponse = _reflection.GeneratedProtocolMessageType('UpdateTrialResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATETRIALRESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.UpdateTrialResponse)
  })
_sym_db.RegisterMessage(UpdateTrialResponse)

EndTrialRequest = _reflection.GeneratedProtocolMessageType('EndTrialRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENDTRIALREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.EndTrialRequest)
  })
_sym_db.RegisterMessage(EndTrialRequest)

EndTrialResponse = _reflection.GeneratedProtocolMessageType('EndTrialResponse', (_message.Message,), {
  'DESCRIPTOR' : _ENDTRIALRESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.EndTrialResponse)
  })
_sym_db.RegisterMessage(EndTrialResponse)

GetBestTrialsRequest = _reflection.GeneratedProtocolMessageType('GetBestTrialsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETBESTTRIALSREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.GetBestTrialsRequest)
  })
_sym_db.RegisterMessage(GetBestTrialsRequest)

GetBestTrialsResponse = _reflection.GeneratedProtocolMessageType('GetBestTrialsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETBESTTRIALSRESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.GetBestTrialsResponse)
  })
_sym_db.RegisterMessage(GetBestTrialsResponse)

GetTrialRequest = _reflection.GeneratedProtocolMessageType('GetTrialRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRIALREQUEST,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.GetTrialRequest)
  })
_sym_db.RegisterMessage(GetTrialRequest)

GetTrialResponse = _reflection.GeneratedProtocolMessageType('GetTrialResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTRIALRESPONSE,
  '__module__' : 'keras_tuner.protos.service_pb2'
  # @@protoc_insertion_point(class_scope:keras_tuner.GetTrialResponse)
  })
_sym_db.RegisterMessage(GetTrialResponse)


_UPDATETRIALREQUEST_METRICSENTRY._options = None

_ORACLE = _descriptor.ServiceDescriptor(
  name='Oracle',
  full_name='keras_tuner.Oracle',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=907,
  serialized_end=1482,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetSpace',
    full_name='keras_tuner.Oracle.GetSpace',
    index=0,
    containing_service=None,
    input_type=_GETSPACEREQUEST,
    output_type=_GETSPACERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateSpace',
    full_name='keras_tuner.Oracle.UpdateSpace',
    index=1,
    containing_service=None,
    input_type=_UPDATESPACEREQUEST,
    output_type=_UPDATESPACERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateTrial',
    full_name='keras_tuner.Oracle.CreateTrial',
    index=2,
    containing_service=None,
    input_type=_CREATETRIALREQUEST,
    output_type=_CREATETRIALRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateTrial',
    full_name='keras_tuner.Oracle.UpdateTrial',
    index=3,
    containing_service=None,
    input_type=_UPDATETRIALREQUEST,
    output_type=_UPDATETRIALRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='EndTrial',
    full_name='keras_tuner.Oracle.EndTrial',
    index=4,
    containing_service=None,
    input_type=_ENDTRIALREQUEST,
    output_type=_ENDTRIALRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetBestTrials',
    full_name='keras_tuner.Oracle.GetBestTrials',
    index=5,
    containing_service=None,
    input_type=_GETBESTTRIALSREQUEST,
    output_type=_GETBESTTRIALSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetTrial',
    full_name='keras_tuner.Oracle.GetTrial',
    index=6,
    containing_service=None,
    input_type=_GETTRIALREQUEST,
    output_type=_GETTRIALRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ORACLE)

DESCRIPTOR.services_by_name['Oracle'] = _ORACLE

# @@protoc_insertion_point(module_scope)



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trial class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hashlib
import random
import time

import tensorflow as tf

from ..protos import keras_tuner_pb2
from . import hyperparameters as hp_module
from . import metrics_tracking
from . import stateful


class TrialStatus:
    RUNNING = "RUNNING"
    IDLE = "IDLE"
    INVALID = "INVALID"
    STOPPED = "STOPPED"
    COMPLETED = "COMPLETED"


class Trial(stateful.Stateful):
    def __init__(self, hyperparameters, trial_id=None, status=TrialStatus.RUNNING):
        self.hyperparameters = hyperparameters
        self.trial_id = generate_trial_id() if trial_id is None else trial_id

        self.metrics = metrics_tracking.MetricsTracker()
        self.score = None
        self.best_step = None
        self.status = status

    def summary(self):
        """Displays a summary of this Trial."""
        print("Trial summary")

        print("Hyperparameters:")
        self.display_hyperparameters()

        if self.score is not None:
            print("Score: {}".format(self.score))

    def display_hyperparameters(self):
        if self.hyperparameters.values:
            for hp, value in self.hyperparameters.values.items():
                print(hp + ":", value)
        else:
            print("default configuration")

    def get_state(self):
        return {
            "trial_id": self.trial_id,
            "hyperparameters": self.hyperparameters.get_config(),
            "metrics": self.metrics.get_config(),
            "score": self.score,
            "best_step": self.best_step,
            "status": self.status,
        }

    def set_state(self, state):
        self.trial_id = state["trial_id"]
        hp = hp_module.HyperParameters.from_config(state["hyperparameters"])
        self.hyperparameters = hp
        self.metrics = metrics_tracking.MetricsTracker.from_config(state["metrics"])
        self.score = state["score"]
        self.best_step = state["best_step"]
        self.status = state["status"]

    @classmethod
    def from_state(cls, state):
        trial = cls(hyperparameters=None)
        trial.set_state(state)
        return trial

    @classmethod
    def load(cls, fname):
        with tf.io.gfile.GFile(fname, "r") as f:
            state_data = f.read()
        return cls.from_state(state_data)

    def to_proto(self):
        if self.score is not None:
            score = keras_tuner_pb2.Trial.Score(
                value=self.score, step=self.best_step
            )
        else:
            score = None
        proto = keras_tuner_pb2.Trial(
            trial_id=self.trial_id,
            hyperparameters=self.hyperparameters.to_proto(),
            score=score,
            status=_convert_trial_status_to_proto(self.status),
            metrics=self.metrics.to_proto(),
        )
        return proto

    @classmethod
    def from_proto(cls, proto):
        instance = cls(
            hp_module.HyperParameters.from_proto(proto.hyperparameters),
            trial_id=proto.trial_id,
            status=_convert_trial_status_to_str(proto.status),
        )
        if proto.HasField("score"):
            instance.score = proto.score.value
            instance.best_step = proto.score.step
        instance.metrics = metrics_tracking.MetricsTracker.from_proto(proto.metrics)
        return instance


def generate_trial_id():
    s = str(time.time()) + str(random.randint(1, int(1e7)))
    return hashlib.sha256(s.encode("utf-8")).hexdigest()[:32]


def _convert_trial_status_to_proto(status):
    ts = keras_tuner_pb2.TrialStatus
    if status is None:
        return ts.UNKNOWN
    elif status == TrialStatus.RUNNING:
        return ts.RUNNING
    elif status == TrialStatus.IDLE:
        return ts.IDLE
    elif status == TrialStatus.INVALID:
        return ts.INVALID
    elif status == TrialStatus.STOPPED:
        return ts.STOPPED
    elif status == TrialStatus.COMPLETED:
        return ts.COMPLETED
    else:
        raise ValueError("Unknown status {}".format(status))


def _convert_trial_status_to_str(status):
    ts = keras_tuner_pb2.TrialStatus
    if status == ts.UNKNOWN:
        return None
    elif status == ts.RUNNING:
        return TrialStatus.RUNNING
    elif status == ts.IDLE:
        return TrialStatus.IDLE
    elif status == ts.INVALID:
        return TrialStatus.INVALID
    elif status == ts.STOPPED:
        return TrialStatus.STOPPED
    elif status == ts.COMPLETED:
        return TrialStatus.COMPLETED
    else:
        raise ValueError("Unknown status {}".format(status))



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Oracle base class."

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import hashlib
import json
import os
import random
import warnings

import tensorflow as tf

from .. import utils
from . import hyperparameters as hp_module
from . import metrics_tracking
from . import stateful
from . import trial as trial_lib

Objective = collections.namedtuple("Objective", "name direction")


class Oracle(stateful.Stateful):
    """Implements a hyperparameter optimization algorithm.

    Args:
        objective: A string or `keras_tuner.Objective` instance. If a string,
            the direction of the optimization (min or max) will be inferred.
        max_trials: Integer, the total number of trials (model configurations)
            to test at most. Note that the oracle may interrupt the search
            before `max_trial` models have been tested if the search space has
            been exhausted.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        seed: Int. Random seed.
    """

    def __init__(
        self,
        objective,
        max_trials=None,
        hyperparameters=None,
        allow_new_entries=True,
        tune_new_entries=True,
        seed=None,
    ):
        self.objective = _format_objective(objective)
        self.max_trials = max_trials
        if not hyperparameters:
            if not tune_new_entries:
                raise ValueError(
                    "If you set `tune_new_entries=False`, you must"
                    "specify the search space via the "
                    "`hyperparameters` argument."
                )
            if not allow_new_entries:
                raise ValueError(
                    "If you set `allow_new_entries=False`, you must"
                    "specify the search space via the "
                    "`hyperparameters` argument."
                )
            self.hyperparameters = hp_module.HyperParameters()
        else:
            self.hyperparameters = hyperparameters
        self.allow_new_entries = allow_new_entries
        self.tune_new_entries = tune_new_entries

        # trial_id -> Trial
        self.trials = {}
        # tuner_id -> Trial
        self.ongoing_trials = {}
        # List of trial_ids in the order of the trials start
        self.start_order = []
        # List of trial_ids in the order of the trials end
        self.end_order = []

        self.seed = seed or random.randint(1, 10000)
        self._seed_state = self.seed
        # Hashes of values tried so far.
        self._tried_so_far = set()
        # Maximum number of identical values that can be generated
        # before we consider the space to be exhausted.
        self._max_collisions = 5

        # Set in `BaseTuner` via `set_project_dir`.
        self.directory = None
        self.project_name = None

        # In multi-worker mode, only the chief of each cluster should report
        # results.
        self.multi_worker = False
        self.should_report = True

    def _populate_space(self, trial_id):
        warnings.warn(
            "The `_populate_space` method is deprecated, "
            "please use `populate_space`.",
            DeprecationWarning,
        )
        return self.populate_space(trial_id)

    def populate_space(self, trial_id):
        """Fill the hyperparameter space with values for a trial.

        This method should be overridden in subclasses and called in
        `create_trial` in order to populate the hyperparameter space with
        values.

        Args:
            trial_id: A string, the ID for this Trial.

        Returns:
            A dictionary with keys "values" and "status", where "values" is
            a mapping of parameter names to suggested values, and "status"
            is the TrialStatus that should be returned for this trial (one
            of "RUNNING", "IDLE", or "STOPPED").
        """
        raise NotImplementedError

    def _score_trial(self, trial):
        warnings.warn(
            "The `_score_trial` method is deprecated, please use `score_trial`.",
            DeprecationWarning,
        )
        self.score_trial(trial)

    def score_trial(self, trial):
        """Score a completed `Trial`.

        This method can be overridden in subclasses to provide a score for
        a set of hyperparameter values. This method is called from `end_trial`
        on completed `Trial`s.

        Args:
            trial: A completed `Trial` object.
        """
        # Assumes single objective, subclasses can override.
        trial.score = trial.metrics.get_best_value(self.objective.name)
        trial.best_step = trial.metrics.get_best_step(self.objective.name)

    def create_trial(self, tuner_id):
        """Create a new `Trial` to be run by the `Tuner`.

        A `Trial` corresponds to a unique set of hyperparameters to be run
        by `Tuner.run_trial`.

        Args:
            tuner_id: A string, the ID that identifies the `Tuner` requesting a
                `Trial`. `Tuners` that should run the same trial (for instance,
                when running a multi-worker model) should have the same ID.

        Returns:
            A `Trial` object containing a set of hyperparameter values to run
            in a `Tuner`.
        """
        # Allow for multi-worker DistributionStrategy within a Trial.
        if tuner_id in self.ongoing_trials:
            return self.ongoing_trials[tuner_id]

        trial_id = trial_lib.generate_trial_id()

        if self.max_trials and len(self.trials) >= self.max_trials:
            status = trial_lib.TrialStatus.STOPPED
            values = None
        else:
            response = self.populate_space(trial_id)
            status = response["status"]
            values = response["values"] if "values" in response else None

        hyperparameters = self.hyperparameters.copy()
        hyperparameters.values = values or {}
        trial = trial_lib.Trial(
            hyperparameters=hyperparameters, trial_id=trial_id, status=status
        )

        if status == trial_lib.TrialStatus.RUNNING:
            self.ongoing_trials[tuner_id] = trial
            self.trials[trial_id] = trial
            self.start_order.append(trial_id)
            self._save_trial(trial)
            self.save()

        return trial

    def update_trial(self, trial_id, metrics, step=0):
        """Used by a worker to report the status of a trial.

        Args:
            trial_id: A string, a previously seen trial id.
            metrics: Dict of float. The current value of this trial's metrics.
            step: Optional float, reporting intermediate results. The current
                value in a timeseries representing the state of the trial. This
                is the value that `metrics` will be associated with.

        Returns:
            Trial object. Trial.status will be set to "STOPPED" if the Trial
            should be stopped early.
        """
        trial = self.trials[trial_id]
        self._check_objective_found(metrics)
        for metric_name, metric_value in metrics.items():
            if not trial.metrics.exists(metric_name):
                direction = _maybe_infer_direction_from_objective(
                    self.objective, metric_name
                )
                trial.metrics.register(metric_name, direction=direction)
            trial.metrics.update(metric_name, metric_value, step=step)
        self._save_trial(trial)
        # To signal early stopping, set Trial.status to "STOPPED".
        return trial.status

    def end_trial(self, trial_id, status="COMPLETED"):
        """Record the measured objective for a set of parameter values.

        Args:
            trial_id: A string, the unique ID for this trial.
            status: A string, one of `"COMPLETED"`, `"INVALID"`. A status of
                `"INVALID"` means a trial has crashed or been deemed
                infeasible.
        """
        trial = None
        for tuner_id, ongoing_trial in self.ongoing_trials.items():
            if ongoing_trial.trial_id == trial_id:
                trial = self.ongoing_trials.pop(tuner_id)
                break

        if not trial:
            raise ValueError("Ongoing trial with id: {} not found.".format(trial_id))

        trial.status = status
        if status == trial_lib.TrialStatus.COMPLETED:
            self.score_trial(trial)
        self.end_order.append(trial_id)
        self._save_trial(trial)
        self.save()

    def get_space(self):
        """Returns the `HyperParameters` search space."""
        return self.hyperparameters.copy()

    def update_space(self, hyperparameters):
        """Add new hyperparameters to the tracking space.

        Already recorded parameters get ignored.

        Args:
            hyperparameters: An updated `HyperParameters` object.
        """
        hps = hyperparameters.space
        new_hps = []
        for hp in hps:
            if not self.hyperparameters._exists(hp.name, hp.conditions):
                new_hps.append(hp)

        if new_hps and not self.allow_new_entries:
            raise RuntimeError(
                "`allow_new_entries` is `False`, but found "
                "new entries {}".format(new_hps)
            )
        if not self.tune_new_entries:
            # New entries should always use the default value.
            return
        self.hyperparameters.merge(new_hps)

    def get_trial(self, trial_id):
        """Returns the `Trial` specified by `trial_id`."""
        return self.trials[trial_id]

    def get_best_trials(self, num_trials=1):
        """Returns the best `Trial`s."""
        trials = [
            t
            for t in self.trials.values()
            if t.status == trial_lib.TrialStatus.COMPLETED
        ]

        sorted_trials = sorted(
            trials,
            key=lambda trial: trial.score,
            # Assumes single objective, subclasses can override.
            reverse=self.objective.direction == "max",
        )
        return sorted_trials[:num_trials]

    def remaining_trials(self):
        if self.max_trials:
            return self.max_trials - len(self.trials.items())
        else:
            return None

    def get_state(self):
        # `self.trials` are saved in their own, Oracle-agnostic files.
        # Just save the IDs for ongoing trials, since these are in `trials`.
        state = {}
        state["ongoing_trials"] = {
            tuner_id: trial.trial_id
            for tuner_id, trial in self.ongoing_trials.items()
        }
        # Hyperparameters are part of the state because they can be added to
        # during the course of the search.
        state["hyperparameters"] = self.hyperparameters.get_config()
        state["seed"] = self.seed
        state["seed_state"] = self._seed_state
        state["tried_so_far"] = list(self._tried_so_far)
        return state

    def set_state(self, state):
        # `self.trials` are saved in their own, Oracle-agnostic files.
        self.ongoing_trials = {
            tuner_id: self.trials[trial_id]
            for tuner_id, trial_id in state["ongoing_trials"].items()
        }
        self.hyperparameters = hp_module.HyperParameters.from_config(
            state["hyperparameters"]
        )
        self.seed = state["seed"]
        self._seed_state = state["seed_state"]
        self._tried_so_far = set(state["tried_so_far"])

    def _set_project_dir(self, directory, project_name, overwrite=False):
        """Sets the project directory and reloads the Oracle."""
        self._directory = directory
        self._project_name = project_name
        if not overwrite and tf.io.gfile.exists(self._get_oracle_fname()):
            tf.get_logger().info(
                "Reloading Oracle from existing project {}".format(
                    self._get_oracle_fname()
                )
            )
            self.reload()

    @property
    def _project_dir(self):
        dirname = os.path.join(str(self._directory), self._project_name)
        utils.create_directory(dirname)
        return dirname

    def save(self):
        # `self.trials` are saved in their own, Oracle-agnostic files.
        super(Oracle, self).save(self._get_oracle_fname())

    def reload(self):
        # Reload trials from their own files.
        trial_fnames = tf.io.gfile.glob(
            os.path.join(self._project_dir, "trial_*", "trial.json")
        )
        for fname in trial_fnames:
            with tf.io.gfile.GFile(fname, "r") as f:
                trial_data = f.read()
            trial_state = json.loads(trial_data)
            trial = trial_lib.Trial.from_state(trial_state)
            self.trials[trial.trial_id] = trial
        try:
            super(Oracle, self).reload(self._get_oracle_fname())
        except KeyError:
            raise RuntimeError(
                "Error reloading `Oracle` from existing project. If you did not "
                "mean to reload from an existing project, change the `project_name` "
                "or pass `overwrite=True` when creating the `Tuner`. Found existing "
                "project at: {}".format(self._project_dir)
            )

    def _get_oracle_fname(self):
        return os.path.join(self._project_dir, "oracle.json")

    def _compute_values_hash(self, values):
        keys = sorted(values.keys())
        s = "".join(str(k) + "=" + str(values[k]) for k in keys)
        return hashlib.sha256(s.encode("utf-8")).hexdigest()[:32]

    def _check_objective_found(self, metrics):
        if isinstance(self.objective, Objective):
            objective_names = [self.objective.name]
        else:
            objective_names = [obj.name for obj in self.objective]
        for metric_name in metrics.keys():
            if metric_name in objective_names:
                objective_names.remove(metric_name)
        if objective_names:
            raise ValueError(
                "Objective value missing in metrics reported to the "
                "Oracle, expected: {}, found: {}".format(
                    objective_names, metrics.keys()
                )
            )

    def _get_trial_dir(self, trial_id):
        dirname = os.path.join(self._project_dir, "trial_" + str(trial_id))
        utils.create_directory(dirname)
        return dirname

    def _save_trial(self, trial):
        # Write trial status to trial directory
        trial_id = trial.trial_id
        trial.save(os.path.join(self._get_trial_dir(trial_id), "trial.json"))

    def _random_values(self):
        """Fills the hyperparameter space with random values.

        Returns:
            A dictionary mapping parameter names to suggested values.
        """
        collisions = 0
        while 1:
            hps = hp_module.HyperParameters()
            # Generate a set of random values.
            for hp in self.hyperparameters.space:
                hps.merge([hp])
                if hps.is_active(hp):  # Only active params in `values`.
                    hps.values[hp.name] = hp.random_sample(self._seed_state)
                    self._seed_state += 1
            values = hps.values
            # Keep trying until the set of values is unique,
            # or until we exit due to too many collisions.
            values_hash = self._compute_values_hash(values)
            if values_hash in self._tried_so_far:
                collisions += 1
                if collisions > self._max_collisions:
                    return None
                continue
            self._tried_so_far.add(values_hash)
            break
        return values


def _format_objective(objective):
    if isinstance(objective, list):
        return [_format_objective(obj) for obj in objective]
    if isinstance(objective, Objective):
        return objective
    if isinstance(objective, str):
        direction = metrics_tracking.infer_metric_direction(objective)
        if direction is None:
            error_msg = (
                'Could not infer optimization direction ("min" or "max") '
                'for unknown metric "{obj}". Please specify the objective  as'
                "a `keras_tuner.Objective`, for example `keras_tuner.Objective("
                '"{obj}", direction="min")`.'
            )
            error_msg = error_msg.format(obj=objective)
            raise ValueError(error_msg)
        return Objective(name=objective, direction=direction)
    else:
        raise ValueError(
            "`objective` not understood, expected str or "
            "`Objective` object, found: {}".format(objective)
        )


def _maybe_infer_direction_from_objective(objective, metric_name):
    if isinstance(objective, Objective):
        objective = [objective]
    for obj in objective:
        if obj.name == metric_name:
            return obj.direction
    return None



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Keras Tuner class."

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import os

import tensorflow as tf
from tensorboard.plugins.hparams import api as hparams_api

from . import base_tuner
from . import hypermodel as hm_module
from . import tuner_utils


class Tuner(base_tuner.BaseTuner):
    """Tuner class for Keras models.

    This is the base `Tuner` class for all tuners for Keras models. It manages
    the building, training, evaluation and saving of the Keras models. New
    tuners can be created by subclassing the class.

    Args:
        oracle: Instance of Oracle class.
        hypermodel: Instance of HyperModel class (or callable that takes
            hyperparameters and returns a Model instance).
        max_model_size: Integer, maximum number of scalars in the parameters of
            a model. Models larger than this are rejected.
        optimizer: Optional `Optimizer` instance.  May be used to override the
            `optimizer` argument in the `compile` step for the models. If the
            hypermodel does not compile the models it generates, then this
            argument must be specified.
        loss: Optional loss. May be used to override the `loss` argument in the
            `compile` step for the models. If the hypermodel does not compile
            the models it generates, then this argument must be specified.
        metrics: Optional metrics. May be used to override the `metrics`
            argument in the `compile` step for the models. If the hypermodel
            does not compile the models it generates, then this argument must
            be specified.
        distribution_strategy: Optional instance of `tf.distribute.Strategy`.
            If specified, each trial will run under this scope. For example,
            `tf.distribute.MirroredStrategy(['/gpu:0', '/gpu:1'])` will run
            each trial on two GPUs. Currently only single-worker strategies are
            supported.
        directory: A string, the relative path to the working directory.
        project_name: A string, the name to use as prefix for files saved by
            this `Tuner`.
        logger: Optional instance of `kerastuner.Logger` class for
            streaming logs for monitoring.
        tuner_id: Optional string, used as the ID of this `Tuner`.
        overwrite: Boolean, defaults to `False`. If `False`, reloads an
            existing project of the same name if one is found. Otherwise,
            overwrites the project.

    Attributes:
        remaining_trials: Number of trials remaining, `None` if `max_trials` is
            not set. This is useful when resuming a previously stopped search.
    """

    def __init__(
        self,
        oracle,
        hypermodel,
        max_model_size=None,
        optimizer=None,
        loss=None,
        metrics=None,
        distribution_strategy=None,
        directory=None,
        project_name=None,
        logger=None,
        tuner_id=None,
        overwrite=False,
    ):

        # Subclasses of `KerasHyperModel` are not automatically wrapped.
        if not isinstance(hypermodel, hm_module.KerasHyperModel):
            hypermodel = hm_module.KerasHyperModel(
                hypermodel,
                max_model_size=max_model_size,
                optimizer=optimizer,
                loss=loss,
                metrics=metrics,
                distribution_strategy=distribution_strategy,
            )

        super(Tuner, self).__init__(
            oracle=oracle,
            hypermodel=hypermodel,
            directory=directory,
            project_name=project_name,
            logger=logger,
            overwrite=overwrite,
        )

        self.distribution_strategy = distribution_strategy

        # Support multi-worker distribution strategies w/ distributed tuning.
        # Only the chief worker in each cluster should report results.
        if self.distribution_strategy is not None and hasattr(
            self.distribution_strategy.extended, "_in_multi_worker_mode"
        ):
            self.oracle.multi_worker = (
                self.distribution_strategy.extended._in_multi_worker_mode()
            )
            self.oracle.should_report = (
                self.distribution_strategy.extended.should_checkpoint
            )

        # Save only the last N checkpoints.
        self._save_n_checkpoints = 10

        self.tuner_id = tuner_id or self.tuner_id

    def _build_and_fit_model(self, trial, fit_args, fit_kwargs):
        """For AutoKeras to override.

        DO NOT REMOVE this function. AutoKeras overrides the function to tune
        tf.data preprocessing pipelines, preprocess the dataset to obtain
        the input shape before building the model, adapt preprocessing layers,
        and tune other fit_args and fit_kwargs.

        Args:
            trial: A `Trial` instance that contains the information needed to
                run this trial. `Hyperparameters` can be accessed via
                `trial.hyperparameters`.
            fit_args: Positional arguments passed by `search`.
            fit_kwargs: Keyword arguments passed by `search`.

        Returns:
            The fit history.
        """
        model = self.hypermodel.build(trial.hyperparameters)
        return model.fit(*fit_args, **fit_kwargs)

    def run_trial(self, trial, *fit_args, **fit_kwargs):
        """Evaluates a set of hyperparameter values.

        This method is called multiple times during `search` to build and
        evaluate the models with different hyperparameters.

        The method is responsible for reporting metrics related to the `Trial`
        to the `Oracle` via `self.oracle.update_trial`.

        Args:
            trial: A `Trial` instance that contains the information needed to
                run this trial. `Hyperparameters` can be accessed via
                `trial.hyperparameters`.
            *fit_args: Positional arguments passed by `search`.
            **fit_kwargs: Keyword arguments passed by `search`.
        """
        # Handle any callbacks passed to `fit`.
        copied_fit_kwargs = copy.copy(fit_kwargs)
        callbacks = fit_kwargs.pop("callbacks", [])
        callbacks = self._deepcopy_callbacks(callbacks)
        self._configure_tensorboard_dir(callbacks, trial)
        # `TunerCallback` calls:
        # - `Tuner.on_epoch_begin`
        # - `Tuner.on_batch_begin`
        # - `Tuner.on_batch_end`
        # - `Tuner.on_epoch_end`
        # These methods report results to the `Oracle` and save the trained Model. If
        # you are subclassing `Tuner` to write a custom training loop, you should
        # make calls to these methods within `run_trial`.
        callbacks.append(tuner_utils.TunerCallback(self, trial))
        copied_fit_kwargs["callbacks"] = callbacks

        self._build_and_fit_model(trial, fit_args, copied_fit_kwargs)

    def save_model(self, trial_id, model, step=0):
        epoch = step
        self._checkpoint_model(model, trial_id, epoch)
        # TODO: save the top epoch checkpoints instead of last ones.
        epoch_to_delete = epoch - self._save_n_checkpoints
        best_epoch = 0
        if epoch > 0:
            # TODO: `get_best_models` would load the `best_step` checkpoint after
            # training. It would break if oracle picks a different `best_step` than
            # `metrics.get_best_step` since it might be deleted due to it was
            # not the `best_epoch` during the training.
            best_epoch = self.oracle.get_trial(trial_id).metrics.get_best_step(
                self.oracle.objective.name
            )
        if epoch > self._save_n_checkpoints and epoch_to_delete != best_epoch:
            self._delete_checkpoint(trial_id, epoch_to_delete)

    def load_model(self, trial):
        model = self.hypermodel.build(trial.hyperparameters)
        # Reload best checkpoint. The Oracle scores the Trial and also
        # indicates at what epoch the best value of the objective was
        # obtained.
        best_epoch = trial.best_step
        with hm_module.maybe_distribute(self.distribution_strategy):
            model.load_weights(
                self._get_checkpoint_fname(trial.trial_id, best_epoch)
            )
        return model

    def on_batch_begin(self, trial, model, batch, logs):
        """Called at the beginning of a batch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            batch: The current batch number within the current epoch.
            logs: Additional metrics.
        """
        pass

    def on_batch_end(self, trial, model, batch, logs=None):
        """Called at the end of a batch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            batch: The current batch number within the current epoch.
            logs: Additional metrics.
        """
        pass

    def on_epoch_begin(self, trial, model, epoch, logs=None):
        """Called at the beginning of an epoch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            epoch: The current epoch number.
            logs: Additional metrics.
        """
        pass

    def on_epoch_end(self, trial, model, epoch, logs=None):
        """Called at the end of an epoch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            epoch: The current epoch number.
            logs: Dict. Metrics for this epoch. This should include
              the value of the objective for this epoch.
        """
        self.save_model(trial.trial_id, model, step=epoch)
        # Report intermediate metrics to the `Oracle`.
        status = self.oracle.update_trial(trial.trial_id, metrics=logs, step=epoch)
        trial.status = status
        if trial.status == "STOPPED":
            model.stop_training = True

    def get_best_models(self, num_models=1):
        """Returns the best model(s), as determined by the tuner's objective.

        The models are loaded with the weights corresponding to
        their best checkpoint (at the end of the best epoch of best trial).

        This method is for querying the models trained during the search.
        For best performance, it is recommended to retrain your Model on the
        full dataset using the best hyperparameters found during `search`,
        which can be obtained using `tuner.get_best_hyperparameters()`.

        Args:
            num_models: Optional number of best models to return.
                Defaults to 1.

        Returns:
            List of trained model instances sorted from the best to the worst.
        """
        # Method only exists in this class for the docstring override.
        return super(Tuner, self).get_best_models(num_models)

    def _deepcopy_callbacks(self, callbacks):
        try:
            callbacks = copy.deepcopy(callbacks)
        except:
            raise ValueError(
                "All callbacks used during a search "
                "should be deep-copyable (since they are "
                "reused across trials). "
                "It is not possible to do `copy.deepcopy(%s)`" % (callbacks,)
            )
        return callbacks

    def _configure_tensorboard_dir(self, callbacks, trial):
        for callback in callbacks:
            if callback.__class__.__name__ == "TensorBoard":
                # Patch TensorBoard log_dir and add HParams KerasCallback
                logdir = self._get_tensorboard_dir(callback.log_dir, trial.trial_id)
                callback.log_dir = logdir
                hparams = tuner_utils.convert_hyperparams_to_hparams(
                    trial.hyperparameters
                )
                callbacks.append(
                    hparams_api.KerasCallback(
                        writer=logdir, hparams=hparams, trial_id=trial.trial_id
                    )
                )

    def _get_tensorboard_dir(self, logdir, trial_id):
        return os.path.join(str(logdir), str(trial_id))

    def _get_checkpoint_dir(self, trial_id, epoch):
        return os.path.join(
            self.get_trial_dir(trial_id), "checkpoints", "epoch_" + str(epoch)
        )

    def _get_checkpoint_fname(self, trial_id, epoch):
        return os.path.join(
            # Each checkpoint is saved in its own directory.
            self._get_checkpoint_dir(trial_id, epoch),
            "checkpoint",
        )

    def _checkpoint_model(self, model, trial_id, epoch):
        fname = self._get_checkpoint_fname(trial_id, epoch)
        # Save in TF format.
        model.save_weights(fname)
        return fname

    def _delete_checkpoint(self, trial_id, epoch):
        tf.io.gfile.rmtree(self._get_checkpoint_dir(trial_id, epoch))



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Tuner that runs multiple executions per Trial."

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import copy
import os

import numpy as np
from tensorboard.plugins.hparams import api as hparams_api
from tensorflow import keras

from . import tuner as tuner_module
from . import tuner_utils


class MultiExecutionTuner(tuner_module.Tuner):
    """A `Tuner` class that averages multiple runs of the process.

    Args:
        oracle: An Oracle instance.
        hypermodel: A `HyperModel` instance (or callable that takes
            hyperparameters and returns a `Model` instance).
        executions_per_trial: Integer, the number of executions (training a
            model from scratch, starting from a new initialization) to run per
            trial (model configuration). Model metrics may vary greatly
            depending on random initialization, hence it is often a good idea
            to run several executions per trial in order to evaluate the
            performance of a given set of hyperparameter values.
        **kwargs: Keyword arguments relevant to all `Tuner` subclasses.
            Please see the docstring for `Tuner`.
    """

    def __init__(self, oracle, hypermodel, executions_per_trial=1, **kwargs):
        super(MultiExecutionTuner, self).__init__(oracle, hypermodel, **kwargs)
        if isinstance(oracle.objective, list):
            raise ValueError(
                "Multi-objective is not supported, found: {}".format(
                    oracle.objective
                )
            )
        self.executions_per_trial = executions_per_trial
        # This is the `step` that will be reported to the Oracle at the end
        # of the Trial. Since intermediate results are not used, this is set
        # to 0.
        self._reported_step = 0

    def on_epoch_end(self, trial, model, epoch, logs=None):
        # Intermediate results are not passed to the Oracle, and
        # checkpointing is handled via a `ModelCheckpoint` callback.
        pass

    def run_trial(self, trial, *fit_args, **fit_kwargs):
        model_checkpoint = keras.callbacks.ModelCheckpoint(
            filepath=self._get_checkpoint_fname(trial.trial_id, self._reported_step),
            monitor=self.oracle.objective.name,
            mode=self.oracle.objective.direction,
            save_best_only=True,
            save_weights_only=True,
        )
        original_callbacks = fit_kwargs.pop("callbacks", [])

        # Run the training process multiple times.
        metrics = collections.defaultdict(list)
        for execution in range(self.executions_per_trial):
            copied_fit_kwargs = copy.copy(fit_kwargs)
            callbacks = self._deepcopy_callbacks(original_callbacks)
            self._configure_tensorboard_dir(callbacks, trial, execution)
            callbacks.append(tuner_utils.TunerCallback(self, trial))
            # Only checkpoint the best epoch across all executions.
            callbacks.append(model_checkpoint)
            copied_fit_kwargs["callbacks"] = callbacks

            history = self._build_and_fit_model(trial, fit_args, copied_fit_kwargs)
            for metric, epoch_values in history.history.items():
                if self.oracle.objective.direction == "min":
                    best_value = np.min(epoch_values)
                else:
                    best_value = np.max(epoch_values)
                metrics[metric].append(best_value)

        # Average the results across executions and send to the Oracle.
        averaged_metrics = {}
        for metric, execution_values in metrics.items():
            averaged_metrics[metric] = np.mean(execution_values)
        self.oracle.update_trial(
            trial.trial_id, metrics=averaged_metrics, step=self._reported_step
        )

    def _configure_tensorboard_dir(self, callbacks, trial, execution=0):
        for callback in callbacks:
            if callback.__class__.__name__ == "TensorBoard":
                # Patch TensorBoard log_dir and add HParams KerasCallback
                logdir = self._get_tensorboard_dir(
                    callback.log_dir, trial.trial_id, execution
                )
                callback.log_dir = logdir
                hparams = tuner_utils.convert_hyperparams_to_hparams(
                    trial.hyperparameters
                )
                callbacks.append(
                    hparams_api.KerasCallback(
                        writer=logdir, hparams=hparams, trial_id=trial.trial_id
                    )
                )

    def _get_tensorboard_dir(self, logdir, trial_id, execution):
        return os.path.join(
            str(logdir), str(trial_id), "execution{}".format(execution)
        )



import random
import sys

import numpy as np
import tensorflow as tf


def test_seed_set_before_each_test():
    """Test that the tests are deterministic, locally and in the CI."""
    assert random.random() == 0.8444218515250481
    assert np.random.uniform() == 0.5488135039273248

    random_tf_value = tf.random.uniform([1], dtype=tf.dtypes.float64)[0].numpy()
    if sys.version_info >= (3, 0):
        assert random_tf_value == 0.3358018290373803
    else:
        assert random_tf_value == 0.3759982263246542



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test mock running KerasTuner in a distributed tuning setting."""

import os
import sys
import time

import pytest
import tensorflow as tf

from . import mock_distribute


@pytest.fixture(scope="module")
def tmp_dir(tmpdir_factory):
    return tmpdir_factory.mktemp("integration_test")


@pytest.mark.skipif(sys.version_info < (3, 0), reason="TODO: Enable test for Py2")
def test_mock_distribute(tmp_dir):
    def process_fn():
        assert "KERASTUNER_ORACLE_IP" in os.environ
        # Wait, to test that other threads aren't overriding env vars.
        time.sleep(1)
        assert isinstance(os.environ, mock_distribute.MockEnvVars)
        tuner_id = os.environ["KERASTUNER_TUNER_ID"]
        if "worker" in tuner_id:
            # Give the chief process time to write its value,
            # as we do not join on the chief since it will run
            # a server.
            time.sleep(2)
        fname = os.path.join(str(tmp_dir), tuner_id)
        with tf.io.gfile.GFile(fname, "w") as f:
            f.write(tuner_id)

    mock_distribute.mock_distribute(process_fn, num_workers=3)

    for tuner_id in {"chief", "worker0", "worker1", "worker2"}:
        fname = os.path.join(str(tmp_dir), tuner_id)
        with tf.io.gfile.GFile(fname, "r") as f:
            assert f.read() == tuner_id


def test_exception_raising():
    def worker_error_fn():
        if "worker" in os.environ["KERASTUNER_TUNER_ID"]:
            raise ValueError("Found a worker error")

    with pytest.raises(ValueError, match="Found a worker error"):
        mock_distribute.mock_distribute(worker_error_fn, num_workers=2)

    def chief_error_fn():
        if "chief" in os.environ["KERASTUNER_TUNER_ID"]:
            raise ValueError("Found a chief error")

    with pytest.raises(ValueError, match="Found a chief error"):
        mock_distribute.mock_distribute(chief_error_fn, num_workers=2)



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import applications
from . import oracles
from . import tuners
from .engine.hypermodel import HyperModel
from .engine.hyperparameters import HyperParameter
from .engine.hyperparameters import HyperParameters
from .engine.logger import CloudLogger
from .engine.logger import Logger
from .engine.oracle import Objective
from .engine.oracle import Oracle
from .engine.tuner import Tuner
from .tuners import BayesianOptimization
from .tuners import Hyperband
from .tuners import RandomSearch
from .tuners import SklearnTuner
from .utils import check_tf_version

check_tf_version()

__version__ = "1.0.4"



import math
import random

import numpy as np
import tensorflow as tf
from scipy import optimize as scipy_optimize

from ..engine import hyperparameters as hp_module
from ..engine import multi_execution_tuner
from ..engine import oracle as oracle_module
from ..engine import trial as trial_lib


def cdist(x, y=None):
    if y is None:
        y = x
    return np.linalg.norm(x[:, None, :] - y[None, :, :], axis=-1)


def solve_triangular(a, b, lower):
    if np.isfinite(a).all() and np.isfinite(b).all():
        a = tf.constant(a, dtype=tf.float32)
        b = tf.constant(b, dtype=tf.float32)
        return tf.linalg.triangular_solve(a, b, lower=lower).numpy()
    else:
        raise ValueError("array must not contain infs or NaNs")


def cho_solve(l_matrix, b):
    # Ax=b LL^T=A => Ly=b L^Tx=y
    y = solve_triangular(l_matrix, b.reshape(-1, 1), lower=True)
    return solve_triangular(l_matrix.T, y.reshape(-1, 1), lower=False)


def matern_kernel(x, y=None):
    # nu = 2.5
    dists = cdist(x, y)
    dists *= math.sqrt(5)
    kernel_matrix = (1.0 + dists + dists ** 2 / 3.0) * np.exp(-dists)
    return kernel_matrix


class GaussianProcessRegressor(object):
    """A Gaussian process regressor.

    Args:
        alpha: Float, the value added to the diagonal of the kernel matrix
            during fitting. It represents the expected amount of noise in the
            observed performances in Bayesian optimization.
        seed: Optional int, the random seed.
    """

    def __init__(self, alpha, seed=None):
        self.kernel = matern_kernel
        self.n_restarts_optimizer = 20
        self.normalize_y = True
        self.alpha = alpha
        self.seed = seed
        self._x = None
        self._y = None

    def fit(self, x, y):
        """Fit the Gaussian process regressor.

        Args:
            x: np.ndarray with shape (samples, features).
            y: np.ndarray with shape (samples,).
        """
        self._x_train = np.copy(x)
        self._y_train = np.copy(y)

        # Normalize y.
        self._y_train_mean = np.mean(self._y_train, axis=0)
        self._y_train_std = np.std(self._y_train, axis=0)
        self._y_train = (self._y_train - self._y_train_mean) / self._y_train_std

        # TODO: choose a theta for the kernel.
        kernel_matrix = self.kernel(self._x_train)
        kernel_matrix[np.diag_indices_from(kernel_matrix)] += self.alpha

        # l_matrix * l_matrix^T == kernel_matrix
        self._l_matrix = np.linalg.cholesky(kernel_matrix)
        self._alpha_vector = cho_solve(self._l_matrix, self._y_train)

    def predict(self, x):
        """Predict the mean and standard deviation of the target.

        Args:
            x: np.ndarray with shape (samples, features).

        Returns:
            Two 1-D vectors, the mean vector and standard deviation vector.
        """
        # Compute the mean.
        kernel_trans = self.kernel(x, self._x_train)
        y_mean = kernel_trans.dot(self._alpha_vector)

        # Compute the variance.
        l_inv = solve_triangular(
            self._l_matrix.T, np.eye(self._l_matrix.shape[0]), lower=False
        )
        kernel_inv = l_inv.dot(l_inv.T)

        y_var = np.ones(len(x), dtype=np.float)
        y_var -= np.einsum(
            "ij,ij->i", np.dot(kernel_trans, kernel_inv), kernel_trans
        )
        y_var[y_var < 0] = 0.0

        # Undo normalize y.
        y_var *= self._y_train_std ** 2
        y_mean = self._y_train_std * y_mean + self._y_train_mean

        return y_mean.flatten(), np.sqrt(y_var)


class BayesianOptimizationOracle(oracle_module.Oracle):
    """Bayesian optimization oracle.

    It uses Bayesian optimization with a underlying Gaussian process model.
    The acquisition function used is upper confidence bound (UCB), which can
    be found [here](
    https://www.cse.wustl.edu/~garnett/cse515t/spring_2015/files/lecture_notes/12.pdf).

    Args:
        objective: A string or `keras_tuner.Objective` instance. If a string,
            the direction of the optimization (min or max) will be inferred.
        max_trials: Integer, the total number of trials (model configurations)
            to test at most. Note that the oracle may interrupt the search
            before `max_trial` models have been tested if the search space has
            been exhausted.
        num_initial_points: Optional number of randomly generated samples as
            initial training data for Bayesian optimization. If left
            unspecified, a value of 3 times the dimensionality of the
            hyperparameter space is used.
        alpha: Float, the value added to the diagonal of the kernel matrix
            during fitting. It represents the expected amount of noise in the
            observed performances in Bayesian optimization. Defaults to 1e-4.
        beta: Float, the balancing factor of exploration and exploitation. The
            larger it is, the more explorative it is. Defaults to 2.6.
        seed: Optional integer, the random seed.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
    """

    def __init__(
        self,
        objective,
        max_trials,
        num_initial_points=None,
        alpha=1e-4,
        beta=2.6,
        seed=None,
        hyperparameters=None,
        allow_new_entries=True,
        tune_new_entries=True,
    ):
        super(BayesianOptimizationOracle, self).__init__(
            objective=objective,
            max_trials=max_trials,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            seed=seed,
        )
        self.num_initial_points = num_initial_points
        self.alpha = alpha
        self.beta = beta
        self.seed = seed or random.randint(1, int(1e4))
        self._seed_state = self.seed
        self._tried_so_far = set()
        self._max_collisions = 20
        self._random_state = np.random.RandomState(self.seed)
        self.gpr = self._make_gpr()

    def _make_gpr(self):
        return GaussianProcessRegressor(
            alpha=self.alpha,
            seed=self.seed,
        )

    def populate_space(self, trial_id):
        # Generate enough samples before training Gaussian process.
        completed_trials = [
            t for t in self.trials.values() if t.status == "COMPLETED"
        ]

        # Use 3 times the dimensionality of the space as the default number of
        # random points.
        dimensions = len(self.hyperparameters.space)
        num_initial_points = self.num_initial_points or 3 * dimensions
        if len(completed_trials) < num_initial_points:
            return self._random_populate_space()

        # Fit a GPR to the completed trials and return the predicted optimum values.
        x, y = self._vectorize_trials()
        try:
            self.gpr.fit(x, y)
        except ValueError as e:
            if "array must not contain infs or NaNs" in str(e):
                return self._random_populate_space()
            raise e

        def _upper_confidence_bound(x):
            x = x.reshape(1, -1)
            mu, sigma = self.gpr.predict(x)
            return mu - self.beta * sigma

        optimal_val = float("inf")
        optimal_x = None
        num_restarts = 50
        bounds = self._get_hp_bounds()
        x_seeds = self._random_state.uniform(
            bounds[:, 0], bounds[:, 1], size=(num_restarts, bounds.shape[0])
        )
        for x_try in x_seeds:
            # Sign of score is flipped when maximizing.
            result = scipy_optimize.minimize(
                _upper_confidence_bound, x0=x_try, bounds=bounds, method="L-BFGS-B"
            )
            if result.fun[0] < optimal_val:
                optimal_val = result.fun[0]
                optimal_x = result.x

        values = self._vector_to_values(optimal_x)
        return {"status": trial_lib.TrialStatus.RUNNING, "values": values}

    def _random_populate_space(self):
        values = self._random_values()
        if values is None:
            return {"status": trial_lib.TrialStatus.STOPPED, "values": None}
        return {"status": trial_lib.TrialStatus.RUNNING, "values": values}

    def get_state(self):
        state = super(BayesianOptimizationOracle, self).get_state()
        state.update(
            {
                "num_initial_points": self.num_initial_points,
                "alpha": self.alpha,
                "beta": self.beta,
            }
        )
        return state

    def set_state(self, state):
        super(BayesianOptimizationOracle, self).set_state(state)
        self.num_initial_points = state["num_initial_points"]
        self.alpha = state["alpha"]
        self.beta = state["beta"]
        self.gpr = self._make_gpr()

    def _vectorize_trials(self):
        x = []
        y = []
        ongoing_trials = set(self.ongoing_trials.values())
        for trial in self.trials.values():
            # Create a vector representation of each Trial's hyperparameters.
            trial_hps = trial.hyperparameters
            vector = []
            for hp in self._nonfixed_space():
                # For hyperparameters not present in the trial (either added after
                # the trial or inactive in the trial), set to default value.
                if (
                    trial_hps.is_active(hp)  # inactive
                    and hp.name in trial_hps.values  # added after the trial
                ):
                    trial_value = trial_hps.values[hp.name]
                else:
                    trial_value = hp.default

                # Embed an HP value into the continuous space [0, 1].
                prob = hp_module.value_to_cumulative_prob(trial_value, hp)
                vector.append(prob)

            if trial in ongoing_trials:
                # "Hallucinate" the results of ongoing trials. This ensures that
                # repeat trials are not selected when running distributed.
                x_h = np.array(vector).reshape((1, -1))
                y_h_mean, y_h_std = self.gpr.predict(x_h)
                # Give a pessimistic estimate of the ongoing trial.
                score = y_h_mean[0] + y_h_std[0]
            elif trial.status == "COMPLETED":
                score = trial.score
                # Always frame the optimization as a minimization for scipy.minimize.
                if self.objective.direction == "max":
                    score = -1 * score
            else:
                continue

            x.append(vector)
            y.append(score)

        x = np.array(x)
        y = np.array(y)
        return x, y

    def _vector_to_values(self, vector):
        hps = hp_module.HyperParameters()
        vector_index = 0
        for hp in self.hyperparameters.space:
            hps.merge([hp])
            if isinstance(hp, hp_module.Fixed):
                value = hp.value
            else:
                prob = vector[vector_index]
                vector_index += 1
                value = hp_module.cumulative_prob_to_value(prob, hp)

            if hps.is_active(hp):
                hps.values[hp.name] = value
        return hps.values

    def _find_closest(self, val, hp):
        values = [hp.min_value]
        while values[-1] + hp.step <= hp.max_value:
            values.append(values[-1] + hp.step)

        array = np.asarray(values)
        index = (np.abs(values - val)).argmin()
        return array[index]

    def _nonfixed_space(self):
        return [
            hp
            for hp in self.hyperparameters.space
            if not isinstance(hp, hp_module.Fixed)
        ]

    def _get_hp_bounds(self):
        bounds = []
        for hp in self._nonfixed_space():
            bounds.append([0, 1])
        return np.array(bounds)


class BayesianOptimization(multi_execution_tuner.MultiExecutionTuner):
    """BayesianOptimization tuning with Gaussian process.

    Args:
        hypermodel: A `HyperModel` instance (or callable that takes
            hyperparameters and returns a Model instance).
        objective: A string or `keras_tuner.Objective` instance. If a string,
            the direction of the optimization (min or max) will be inferred.
        max_trials: Integer, the total number of trials (model configurations)
            to test at most. Note that the oracle may interrupt the search
            before `max_trial` models have been tested if the search space has
            been exhausted.
        num_initial_points: Optional number of randomly generated samples as
            initial training data for Bayesian optimization. If left
            unspecified, a value of 3 times the dimensionality of the
            hyperparameter space is used.
        alpha: Float, the value added to the diagonal of the kernel matrix
            during fitting. It represents the expected amount of noise in the
            observed performances in Bayesian optimization. Defaults to 1e-4.
        beta: Float, the balancing factor of exploration and exploitation. The
            larger it is, the more explorative it is. Defaults to 2.6.
        seed: Optional integer, the random seed.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        **kwargs: Keyword arguments relevant to all `Tuner` subclasses. Please
            see the docstring for `Tuner`.
    """

    def __init__(
        self,
        hypermodel,
        objective,
        max_trials,
        num_initial_points=2,
        alpha=1e-4,
        beta=2.6,
        seed=None,
        hyperparameters=None,
        tune_new_entries=True,
        allow_new_entries=True,
        **kwargs
    ):
        oracle = BayesianOptimizationOracle(
            objective=objective,
            max_trials=max_trials,
            num_initial_points=num_initial_points,
            alpha=alpha,
            beta=beta,
            seed=seed,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
        )
        super(
            BayesianOptimization,
            self,
        ).__init__(oracle=oracle, hypermodel=hypermodel, **kwargs)



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mock running KerasTuner in a distributed tuning setting."""

import os
import sys
import threading
from unittest import mock

import portpicker
import six


class ExceptionStoringThread(threading.Thread):
    def run(self):
        self.raised_exception = None
        try:
            super(ExceptionStoringThread, self).run()
        except BaseException:
            self.raised_exception = sys.exc_info()


class MockEnvVars(dict):
    """Allows setting different environment variables in threads."""

    def __init__(self):
        self.thread_local = threading.local()
        self.initial_env_vars = os.environ.copy()

    def _setup_thread(self):
        if getattr(self.thread_local, "environ", None) is None:
            self.thread_local.environ = self.initial_env_vars.copy()

    def get(self, name, default=None):
        self._setup_thread()
        return self.thread_local.environ.get(name, default)

    def __setitem__(self, name, value):
        self._setup_thread()
        self.thread_local.environ[name] = value

    def __getitem__(self, name):
        self._setup_thread()
        return self.thread_local.environ[name]

    def __contains__(self, name):
        self._setup_thread()
        return name in self.thread_local.environ


def mock_distribute(fn, num_workers=2):
    """Runs `fn` in multiple processes, setting appropriate env vars."""
    port = str(portpicker.pick_unused_port())
    with mock.patch.object(os, "environ", MockEnvVars()):

        def chief_fn():
            # The IP address of the chief Oracle. Run in distributed mode when
            # present. Cloud oracle does not run in this mode because the Cloud
            # API coordinates workers itself.
            os.environ["KERASTUNER_ORACLE_IP"] = "127.0.0.1"
            # The port of the chief Oracle.
            os.environ["KERASTUNER_ORACLE_PORT"] = port
            # The ID of this process. 'chief' will run the OracleServicer server.
            os.environ["KERASTUNER_TUNER_ID"] = "chief"
            fn()

        chief_thread = ExceptionStoringThread(target=chief_fn)
        chief_thread.daemon = True
        chief_thread.start()

        worker_threads = []
        for i in range(num_workers):

            def worker_fn():
                os.environ["KERASTUNER_ORACLE_IP"] = "127.0.0.1"
                os.environ["KERASTUNER_ORACLE_PORT"] = port
                # Workers that are part of the same multi-worker
                # DistributionStrategy should have the same TUNER_ID.
                os.environ["KERASTUNER_TUNER_ID"] = "worker{}".format(i)
                fn()

            worker_thread = ExceptionStoringThread(target=worker_fn)
            worker_thread.start()
            worker_threads.append(worker_thread)

        for worker_thread in worker_threads:
            worker_thread.join()

        if chief_thread.raised_exception:
            six.reraise(*chief_thread.raised_exception)
        for worker_thread in worker_threads:
            if worker_thread.raised_exception is not None:
                six.reraise(*worker_thread.raised_exception)



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Tuner class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import time

import numpy as np
import six
import tensorflow as tf
from tensorboard.plugins.hparams import api as hparams_api
from tensorflow import keras

from .. import utils
from . import hyperparameters as hp_module


class TunerStats(object):
    """Track tuner statistics."""

    def __init__(self):
        self.num_generated_models = 0  # overall number of instances generated
        self.num_invalid_models = 0  # how many models didn't work
        self.num_oversized_models = 0  # num models with params> max_params

    def summary(self, extended=False):
        print("Tuning stats")
        for setting, value in self.get_config():
            print(setting + ":", value)

    def get_config(self):
        return {
            "num_generated_models": self.num_generated_models,
            "num_invalid_models": self.num_invalid_models,
            "num_oversized_models": self.num_oversized_models,
        }

    @classmethod
    def from_config(cls, config):
        stats = cls()
        stats.num_generated_models = config["num_generated_models"]
        stats.num_invalid_models = config["num_invalid_models"]
        stats.num_oversized_models = config["num_oversized_models"]
        return stats


def get_max_epochs_and_steps(fit_args, fit_kwargs):
    if fit_args:
        x = tf.nest.flatten(fit_args)[0]
    else:
        x = tf.nest.flatten(fit_kwargs.get("x"))[0]
    batch_size = fit_kwargs.get("batch_size", 32)
    if hasattr(x, "__len__"):
        max_steps = math.ceil(float(len(x)) / batch_size)
    else:
        max_steps = fit_kwargs.get("steps")
    max_epochs = fit_kwargs.get("epochs", 1)
    return max_epochs, max_steps


class TunerCallback(keras.callbacks.Callback):
    def __init__(self, tuner, trial):
        super(TunerCallback, self).__init__()
        self.tuner = tuner
        self.trial = trial

    def on_epoch_begin(self, epoch, logs=None):
        self.tuner.on_epoch_begin(self.trial, self.model, epoch, logs=logs)

    def on_batch_begin(self, batch, logs=None):
        self.tuner.on_batch_begin(self.trial, self.model, batch, logs)

    def on_batch_end(self, batch, logs=None):
        self.tuner.on_batch_end(self.trial, self.model, batch, logs)

    def on_epoch_end(self, epoch, logs=None):
        self.tuner.on_epoch_end(self.trial, self.model, epoch, logs=logs)


# TODO: Add more extensive display.
class Display(object):
    def __init__(self, oracle, verbose=1):
        self.verbose = verbose
        self.oracle = oracle
        self.trial_number = 0
        self.col_width = 18

        # Start time for the overall search
        self.search_start = None

        # Start time of the latest trial
        self.trial_start = None

    def on_trial_begin(self, trial):
        if self.verbose >= 1:

            self.trial_number += 1
            print()
            print("Search: Running Trial #{}".format(self.trial_number))
            print()

            self.trial_start = time.time()
            if self.search_start is None:
                self.search_start = time.time()

            self.show_hyperparameter_table(trial)
            print()

    def on_trial_end(self, trial):
        if self.verbose >= 1:
            utils.try_clear()

            time_taken_str = self.format_time(time.time() - self.trial_start)
            print("Trial {} Complete [{}]".format(self.trial_number, time_taken_str))

            if trial.score is not None:
                print("{}: {}".format(self.oracle.objective.name, trial.score))

            print()
            best_trials = self.oracle.get_best_trials()
            if len(best_trials) > 0:
                best_score = best_trials[0].score
            else:
                best_score = None
            print(
                "Best {} So Far: {}".format(self.oracle.objective.name, best_score)
            )

            time_elapsed_str = self.format_time(time.time() - self.search_start)
            print("Total elapsed time: {}".format(time_elapsed_str))

    def show_hyperparameter_table(self, trial):
        template = "{{0:{0}}}|{{1:{0}}}|{{2:{0}}}".format(self.col_width)
        best_trials = self.oracle.get_best_trials()
        if len(best_trials) > 0:
            best_trial = best_trials[0]
        else:
            best_trial = None
        if trial.hyperparameters.values:
            print(template.format("Hyperparameter", "Value", "Best Value So Far"))
            for hp, value in trial.hyperparameters.values.items():
                if best_trial:
                    best_value = best_trial.hyperparameters.values.get(hp)
                else:
                    best_value = "?"
                print(
                    template.format(
                        self.format_value(hp),
                        self.format_value(value),
                        self.format_value(best_value),
                    )
                )
        else:
            print("default configuration")

    def format_value(self, val):
        if isinstance(val, (int, float)) and not isinstance(val, bool):
            return "{:.5g}".format(val)
        else:
            val_str = str(val)
            if len(val_str) > self.col_width:
                val_str = val_str[: self.col_width - 3] + "..."
            return val_str

    def format_time(self, t):
        return time.strftime("%Hh %Mm %Ss", time.gmtime(t))


def average_histories(histories):
    """Averages the per-epoch metrics from multiple executions."""
    averaged = {}
    metrics = histories[0].keys()
    for metric in metrics:
        values = []
        for epoch_values in six.moves.zip_longest(
            *[h[metric] for h in histories], fillvalue=np.nan
        ):
            values.append(np.nanmean(epoch_values))
        averaged[metric] = values
    # Convert {str: [float]} to [{str: float}]
    averaged = [dict(zip(metrics, vals)) for vals in zip(*averaged.values())]
    return averaged


def convert_hyperparams_to_hparams(hyperparams):
    """Converts KerasTuner HyperParameters to TensorBoard HParams."""
    hparams = {}
    for hp in hyperparams.space:
        hparams_value = {}
        try:
            hparams_value = hyperparams.get(hp.name)
        except ValueError:
            continue

        hparams_domain = {}
        if isinstance(hp, hp_module.Choice):
            hparams_domain = hparams_api.Discrete(hp.values)
        elif isinstance(hp, hp_module.Int):
            if hp.step is not None and hp.step != 1:
                # Note: `hp.max_value` is inclusive, unlike the end index
                # of Python `range()`, which is exclusive
                values = list(range(hp.min_value, hp.max_value + 1, hp.step))
                hparams_domain = hparams_api.Discrete(values)
            else:
                hparams_domain = hparams_api.IntInterval(hp.min_value, hp.max_value)
        elif isinstance(hp, hp_module.Float):
            if hp.step is not None:
                # Note: `hp.max_value` is inclusive, unlike the end index
                # of Numpy's arange(), which is exclusive
                values = np.arange(
                    hp.min_value, hp.max_value + 1e-7, step=hp.step
                ).tolist()
                hparams_domain = hparams_api.Discrete(values)
            else:
                hparams_domain = hparams_api.RealInterval(hp.min_value, hp.max_value)
        elif isinstance(hp, hp_module.Boolean):
            hparams_domain = hparams_api.Discrete([True, False])
        elif isinstance(hp, hp_module.Fixed):
            hparams_domain = hparams_api.Discrete([hp.value])
        else:
            raise ValueError("`HyperParameter` type not recognized: {}".format(hp))

        hparams_key = hparams_api.HParam(hp.name, hparams_domain)
        hparams[hparams_key] = hparams_value

    return hparams



import pytest


def test_kerastuner_same_as_keras_tuner():
    with pytest.deprecated_call():
        import kerastuner
        from kerastuner.tuners import RandomSearch
        from kerastuner.tuners import BayesianOptimization
        from kerastuner.tuners import Hyperband
        from kerastuner.tuners import Sklearn  # noqa: F401
        from kerastuner.oracles import RandomSearch  # noqa: F401,F811
        from kerastuner.oracles import BayesianOptimization  # noqa: F401,F811
        from kerastuner.oracles import Hyperband  # noqa: F401,F811
        from kerastuner.engine.base_tuner import BaseTuner  # noqa: F401
        from kerastuner.engine.conditions import Condition  # noqa: F401
        from kerastuner.engine.hypermodel import HyperModel  # noqa: F401
        from kerastuner.engine.hyperparameters import HyperParameter  # noqa: F401
        from kerastuner.engine.hyperparameters import HyperParameters  # noqa: F401
        from kerastuner.engine.logger import CloudLogger  # noqa: F401
        from kerastuner.engine.logger import Logger  # noqa: F401
        from kerastuner.engine.metrics_tracking import (  # noqa: F401
            MetricObservation,
        )
        from kerastuner.engine.oracle import Objective  # noqa: F401
        from kerastuner.engine.oracle import Oracle  # noqa: F401
        from kerastuner.engine.tuner import Tuner  # noqa: F401
        from kerastuner.engine.stateful import Stateful  # noqa: F401
        from kerastuner.engine.trial import Trial  # noqa: F401
        from kerastuner.engine.multi_execution_tuner import (  # noqa: F401
            MultiExecutionTuner,
        )
        from kerastuner.applications import HyperResNet  # noqa: F401
        from kerastuner.applications import HyperXception  # noqa: F401

    import keras_tuner

    attr1 = [attr for attr in dir(kerastuner) if not attr.startswith("__")]
    attr2 = [attr for attr in dir(keras_tuner) if not attr.startswith("__")]

    assert len(attr1) > 20
    assert set(attr1) >= set(attr2)



# Copyright 2019 The Keras Tuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script."""

from __future__ import absolute_import

from setuptools import find_packages
from setuptools import setup

version = "1.0.4"

setup(
    name="keras-tuner",
    version=version,
    description="Hypertuner for Keras",
    url="https://github.com/keras-team/keras-tuner",
    author="The Keras Tuner authors",
    author_email="kerastuner@google.com",
    license="Apache License 2.0",
    # tensorflow isn't a dependency because it would force the
    # download of the gpu version or the cpu version.
    # users should install it manually.
    install_requires=[
        "packaging",
        "numpy",
        "requests",
        "scipy",
        "tensorboard",
        "ipython",
        "kt-legacy",
    ],
    extras_require={
        "tests": [
            "pytest",
            "flake8",
            "isort",
            "black",
            "pandas",
            "portpicker",
            "pytest-xdist",
            "pytest-cov",
            "scikit-learn",
        ],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
    packages=find_packages(exclude=("tests",)),
)
